from functools import wraps
import pyvista
from .plot import reset_plotter

__all__ = ["set_interactive", "start_xvfb"]


# Expose pyvista callers
@wraps(pyvista.start_xvfb)
def start_xvfb(*args, **kwargs):
    pyvista.start_xvfb(*args, **kwargs)


start_xvfb.__doc__ = start_xvfb.__doc__.split("Notes", 1)[
    0
]  # Remove pyvista specific examples


def set_interactive(flag: bool = True, use_xvfb: bool = True, xvfb_wait: float = 0.1):
    """
    Change settings to swith from interactive to non-interactive plots.

    If `set` is `False`, this runs:

    >>> pyvista.OFF_SCREEN = False
    >>> pyvista4dolfinx.reset_plotter()

    If `set` and `use_xvfb` are `True` (this requires libgl1 libglx-mesa0
    and xvfb to be installed on a unix system):

    >>> pyvista.OFF_SCREEN = True
    >>> pyvista.start_xvfb()
    >>> pyvista4dolfinx.reset_plotter()

    If `set` is `True` and `use_xvfb` is `False` (this requires a renderable
    screen to be configured. May be tricky to set up properly):

    >>> pyvista.OFF_SCREEN = True
    >>> pyvista4dolfinx.reset_plotter()

    Parameters
    ----------
    set : bool
        Whether to set interactive plots, default True.
    use_xvfb : bool
        Whether to call `start_xvfb` in non-interactive mode, default True
    xvfb_wait : float
        How long to wait while setting up the virtual frame buffer.
        pyvista's default is 3 seconds, which we reduced to 0.1 second.
    """
    if flag:
        pyvista.OFF_SCREEN = False
        reset_plotter()
    else:
        pyvista.OFF_SCREEN = True
        if use_xvfb:
            start_xvfb(wait=xvfb_wait)
        reset_plotter()



# Copyright (C) 2025 Stein K.F. Stoter
#
# This file is part of pyvista4dolfinx
#
# Credit to Jørgen S. Dokken and Garth N. Wells
#
# SPDX-License-Identifier:    MIT

import dolfinx, pyvista, numpy as np


def _gather_grid(
    mesh: dolfinx.mesh.Mesh,
    u: dolfinx.fem.Function | None = None,
    warp: dolfinx.fem.Function | bool = False,
    name: str = "",
    dim: int | None = None,
    entities: np.typing.NDArray | None = None,
) -> pyvista.UnstructuredGrid | None:
    """
    Creates an unstructured grid filled with field values associated to
    `u`, warped by the `warp` vector, obtained from gathering the data from
    all processes onto rank 0.

    Parameters
    ----------
    mesh : dolfinx.mesh.Mesh
        Mesh associated to `u` and `warp`.
    u : dolfinx.fem.Function | None, optional
        Field of which to obtain the grid, by default None
    warp: dolfinx.fem.Function | bool, optional
        Field by which to warp the mesh, by default False
    name: str, optional
        Which name to give the data in the mesh. If empty string, the name
        of the `u` field is used. By default ""
    dim: int | None, optional
        Dimension of the grid to be extracted (e.g., for obtaining a
        wireframe for a 2D mesh). None means the topological dimension. By
        default None
    entities: np.typing.NDArray | None, optional
        Which entities to extract from the entire mesh, for plotting mesh
        segments. None means all. By default None

    Returns
    -------
    pyvista.UnstructuredGrid | None
        Only returns the `pyvista.UnstructuredGrid` object on
        `MPI.COMM_WORLD.rank == 0`, else `None`.
    """
    # See: https://fenicsproject.discourse.group/t/how-to-collect-the-global-mesh-without-writing-a-file/12445/3

    # Correct objects on which to operate for given input
    vtk_mesh_from = (
        mesh
        if u is None and type(warp) == bool
        else u.function_space if u is not None else warp.function_space
    )
    index_map_plottable = (
        vtk_mesh_from.geometry.index_map()
        if vtk_mesh_from is mesh
        else vtk_mesh_from.dofmap.index_map
    )
    dim_ = mesh.topology.dim if dim is None else dim
    if not name and u is not None:
        name = u.name

    # Create local VTK mesh data structures)
    topology, cell_types, geometry = (
        np.empty((0,), dtype=np.int32),
        np.empty((0,)),
        np.empty((0, 3)),
    )
    if entities is None or len(entities) > 0:
        args = (
            [vtk_mesh_from, dim, entities] if vtk_mesh_from is mesh else [vtk_mesh_from]
        )
        topology, cell_types, geometry = dolfinx.plot.vtk_mesh(*args)
        num_cells_local = mesh.topology.index_map(dim_).size_local
        num_cells_local_geom = index_map_plottable.size_local

        # Restructure topology vector in case of parallel execution
        num_dofs_per_cell = topology[0]
        topology_dofs = (np.arange(len(topology)) % (num_dofs_per_cell + 1)) != 0
        global_dofs = index_map_plottable.local_to_global(
            topology[topology_dofs].copy()
        )
        topology[topology_dofs] = global_dofs

        # Final data shapes
        topology = topology[: (num_dofs_per_cell + 1) * num_cells_local]
        geometry = geometry[:num_cells_local_geom, :]
        cell_types = cell_types[:num_cells_local]

    # Gather mesh data
    global_topology = mesh.comm.gather(topology, root=0)
    global_geometry = mesh.comm.gather(geometry, root=0)
    global_cell_types = mesh.comm.gather(cell_types, root=0)

    # Gather function data
    if u is not None:
        V = u.function_space
        u_dim = u.ufl_shape[0] if len(u.ufl_shape) > 0 else 1
        num_dofs_local = V.dofmap.index_map.size_local * V.dofmap.index_map_bs
        global_vals = mesh.comm.gather(u.x.array[:num_dofs_local], root=0)

    # Gather warp function data
    if type(warp) is not bool:
        V_warp = warp.function_space
        warp_dim = warp.ufl_shape[0] if len(warp.ufl_shape) > 0 else 1
        num_dofs_local_warp = (
            V_warp.dofmap.index_map.size_local * V_warp.dofmap.index_map_bs
        )
        global_warp = mesh.comm.gather(warp.x.array[:num_dofs_local_warp], root=0)

    if mesh.comm.rank == 0:
        # Stack data
        root_geom = np.vstack(global_geometry)
        root_top = np.concatenate(global_topology)
        root_ct = np.concatenate(global_cell_types)

        # Create and fill the unstructured grid
        grid = pyvista.UnstructuredGrid(root_top, root_ct, root_geom)

        # Add u field to grid
        if u is not None:
            # Restructure values vector
            root_vals = np.concatenate(global_vals)
            num_dofs = root_geom.shape[0]
            values = np.zeros((num_dofs, 3), dtype=np.float64)
            values[:, :u_dim] = root_vals.reshape(num_dofs, V.dofmap.index_map_bs)
            grid.point_data[name] = values
            grid.set_active_scalars(name)

        # Add warp field to grid
        if type(warp) is not bool:
            root_warp = np.concatenate(global_warp)
            num_dofs = len(root_warp) // V_warp.dofmap.index_map_bs
            values_warp = np.zeros((num_dofs, 3), dtype=np.float64)
            values_warp[:, :warp_dim] = root_warp.reshape(num_dofs, warp_dim)
            grid.point_data["warp"] = values_warp

        # Warp the grid
        if type(warp) is not bool:
            grid = grid.warp_by_vector("warp")
        elif warp:
            if u_dim == mesh.geometry.dim:
                grid = grid.warp_by_vector(name)
            elif u_dim == 1 and mesh.geometry.dim in [1, 2]:
                grid = grid.warp_by_scalar(name)
            else:
                raise NotImplementedError

        return grid


def _gather_meshtags(
    meshtags: dolfinx.mesh.MeshTags, mesh: dolfinx.mesh.Mesh, tag: int | None = None
) -> list[np.typing.NDArray, np.typing.NDArray]:
    """
    Gathers the values of a `MeshTags` object on process 0. Returns the
    rank-local indices of each process and the gathered values on process 0.

    Parameters
    ----------
    meshtags : dolfinx.mesh.MeshTags
        The meshtags object
    mesh : dolfinx.mesh.Mesh
        The mesh that this meshtags object refers to
    tag : int | None, optional
        Which tag to extract. `None` means all. By default None

    Returns
    -------
    list[np.typing.NDArray, np.typing.NDArray]
        [local_indices, root_values].
        root_values is `None` on all MPI ranks other than 0.
    """
    local_indices = np.copy(meshtags.indices)
    local_values = np.copy(meshtags.values)

    # Mask non-tagged and ghost
    mask = np.full_like(local_indices, True, dtype=np.bool_)
    index_map = mesh.topology.index_map(meshtags.dim)
    local_ghost_indices = index_map.global_to_local(index_map.ghosts)
    ghosts_ind = np.isin(local_indices, local_ghost_indices)
    mask[ghosts_ind] = False
    if tag:
        mask[local_values != tag] = False
    local_indices = local_indices[mask]
    local_values = local_values[mask]

    gathered_values = mesh.comm.gather(local_values, root=0)
    if mesh.comm.rank == 0:
        root_values = np.concatenate(gathered_values)
        return local_indices, root_values
    return local_indices, None



# Copyright (C) 2025 Stein K.F. Stoter
#
# This file is part of pyvista4dolfinx
#
# SPDX-License-Identifier:    MIT

from functools import singledispatch, wraps
from typing import Any
import atexit

import basix
import dolfinx
from mpi4py import MPI
import numpy as np
import pyvista
import ufl.measure

from .safeplotter import Plotter
from .gather import _gather_grid, _gather_meshtags

__all__ = [
    "show",
    "screenshot",
    "reset_plotter",
    "plot",
    "plot_function_scalar",
    "plot_function_vector",
    "plot_mesh",
    "plot_meshtags",
]

pyvista.global_theme.allow_empty_mesh = True


@wraps(Plotter.show)
def show(*args, auto_close: bool = None, **kwargs):
    global PLOTTER
    ret = PLOTTER.show(*args, auto_close=auto_close, **kwargs)
    if auto_close is not False:
        reset_plotter()
    return ret


_show = show
show.__name__ = "show"
show.__qualname__ = "show"
show.__doc__ = show.__doc__.split("Examples", 1)[0]  # Remove pyvista specific examples


@wraps(Plotter.screenshot)
def screenshot(filename: str, *args, **kwargs):
    global PLOTTER
    ret = PLOTTER.screenshot(filename, *args, **kwargs)
    reset_plotter()
    return ret


screenshot.__name__ = "screenshot"
screenshot.__qualname__ = "screenshot"
screenshot.__doc__ = screenshot.__doc__.split("Examples", 1)[
    0
]  # Remove pyvista specific examples


# Module global plotter object on rank 0. Empty plotter on all other ranks.
PLOTTER = Plotter()


def reset_plotter(*args, plotter: Plotter = None, **kwargs):
    """
    Rsets the default module global `PLOTTER`. If specified, it is set to
    `plotter`. Else, a new empty `Plotter` is created and arguments and key-word
    arguments passed, are forwarded to the initialization of the new `Plotter`.

    Parameters
    ----------
    plotter : Plotter, optional
        If specified, the module global PLOTTER is set to this plotter,
        by default None
    """
    global PLOTTER
    PLOTTER = plotter if plotter is not None else Plotter(*args, **kwargs)


atexit.register(
    reset_plotter
)  # Mitigate python exiting conflict due to wrong PLOTTER state.


@singledispatch
def plot(plottable: Any, *args, **kwargs) -> Plotter:
    """
    Plots the plottable and returns the pyvista `Plotter` object. Depending
    on the type of the plottable, the call is dispatched to:

    >>> pyvista4dolfinx.plot.plot_function_scalar(plottable, *args, **kwargs)  # for scalar dolfinx.fem.Function
    >>> pyvista4dolfinx.plot.plot_function_vector(plottable, *args, **kwargs)  # for vector dolfinx.fem.Function
    >>> pyvista4dolfinx.plot.plot_mesh(plottable, *args, **kwargs)  # for dolfinx.mesh.Mesh
    >>> pyvista4dolfinx.plot.plot_meshtags(plottable, *args, **kwargs)  # for dolfinx.mesh.MeshTags
    >>> pyvista4dolfinx.plot.plot_measure(plottable, *args, **kwargs)  # for ufl.measure.Measure

    Parameters
    ----------
    plottable : dolfinx.fem.Function | dolfinx.mesh.Mesh | dolfinx.mesh.MeshTags
        The plottable entity.

    Returns
    -------
    Plotter
        An mpi-safe child of pyvista.Plotter.
    """
    raise NotImplementedError(
        f"Plotting objects of type {type(plottable)} not implemented"
    )


@plot.register
def plot_function_(plottable: dolfinx.fem.Function, *args, **kwargs) -> Plotter:
    if len(plottable.ufl_shape) > 0:
        return plot_function_vector(plottable, *args, **kwargs)
    return plot_function_scalar(plottable, *args, **kwargs)


@plot.register
def plot_mesh_(
    plottable: dolfinx.mesh.Mesh,
    *args,
    show_partitioning: bool = False,
    **kwargs,
) -> Plotter:
    if show_partitioning:
        meshtags = _get_partitioning_meshtags(plottable)
        return plot_meshtags(meshtags, *args, mesh=plottable, name="MPI rank", **kwargs)
    return plot_mesh(plottable, *args, **kwargs)


@plot.register
def plot_meshtags_(plottable: dolfinx.mesh.MeshTags, *args, **kwargs) -> Plotter:
    return plot_meshtags(plottable, *args, **kwargs)


@plot.register
def plot_measure_(plottable: ufl.measure.Measure, *args, **kwargs) -> Plotter:
    return plot_measure(plottable, *args, **kwargs)


def plot_function_scalar(
    u: dolfinx.fem.Function,
    warp: dolfinx.fem.Function | bool = False,
    show_mesh: bool = True,
    name: str = "",
    plotter: Plotter | None = None,
    clear_plotter: bool = False,
    subplot: tuple | None = None,
    show: bool = False,
    **kwargs,
) -> Plotter:
    """
    Produce a contourplot of scalar-valued function `u`.

    Parameters
    ----------
    u : dolfinx.fem.Function
        The plotted scalar-valued function.
    warp: dolfinx.fem.Function | bool, optional
        Vector-valued function by which to warp the mesh. If `True`, it is
        assumed this is a 2D plot and the mesh is warped by `u` in
        `z`-direction, by default False
    show_mesh: bool, optional
        Whether to overlay a plot of the mesh edges.
    name : str
        The name to give field `u` in the colorbar.
    plotter: Plotter | None, optional
        The Plotting object to which to add this data. If None, the default
        (module-level) instance is used, by default None
    clear_plotter : bool, optional
        Whether to clear/reset the default Plotter instance, by default
        False
    subplot : tuple | None, optional
        If the Plotter is initialized with multiple subplots, specify with
        a tuple which sublot to plot in, by default None
    show : bool, optional
        Whether to show upon completion. Alternatively, call `.show()` on
        the returned Plotter `instance`. By default False
    **kwargs
        The remaining keyword arguments are passed to the
        `plotter.add_mesh` call.

    Returns
    -------
    Plotter
        An mpi-safe child of pyvista.Plotter.
    """

    global PLOTTER
    if clear_plotter:
        reset_plotter()
    if subplot is not None:
        PLOTTER.subplot(*subplot)

    u, warp = _compatible_scalar_u_warp(u, warp)

    # Gather data onto MPI process 0
    grid = _gather_grid(u.function_space.mesh, u, warp=warp, name=name)
    if MPI.COMM_WORLD.rank != 0:
        return PLOTTER

    # Set the global plotter
    PLOTTER = plotter if plotter is not None else PLOTTER

    # Default plotting options
    defaults = {"show_scalar_bar": True}

    # Visualize
    options = defaults | kwargs
    PLOTTER.add_mesh(grid, **options)
    if show_mesh:
        _plot_feature_edges(PLOTTER, grid)
    if warp is not True and u.function_space.mesh.geometry.dim in [1, 2]:
        PLOTTER.view_xy()

    if show:
        _show()

    return PLOTTER


def plot_function_vector(
    u: dolfinx.fem.Function,
    warp: dolfinx.fem.Function | bool = False,
    show_mesh: bool = True,
    name: str = "",
    plotter: pyvista.Plotter | None = None,
    clear_plotter: bool = False,
    subplot: tuple | None = None,
    show: bool = False,
    **kwargs,
) -> Plotter:
    """
    Produce a glyph plot of `u`.

    Parameters
    ----------
    u : dolfinx.fem.Function.
        The plotted vector-valued function.
    warp: dolfinx.fem.Function | bool, optional
        Vector-valued function by which to warp the mesh. If
        `True`, mesh is warped by `u`, by default False
    show_mesh: bool, optional
        Whether to overlay a plot of the mesh edges.
    name : str
        The name to give field `u` in the colorbar.
    show_mesh : bool, optional
        Whether to also show the mesh, by default True
    plotter: Plotter | None, optional
        The Plotting object to which to add this data. If None, the default
        (module-level) instance is used, by default None
    clear_plotter : bool, optional
        Whether to clear/reset the default Plotter instance, by default
        False
    subplot : tuple | None, optional
        If the Plotter is initialized with multiple subplots, specify with
        a tuple which sublot to plot in, by default None
    show : bool, optional
        Whether to show upon completion. Alternatively, call `.show()` on
        the returned Plotter `instance`. By default False
    **kwargs
        The remaining keyword arguments are passed to the
        `plotter.add_mesh` call.

    Returns
    -------
    Plotter
        An mpi-safe child of pyvista.Plotter.
    """
    global PLOTTER
    if clear_plotter:
        reset_plotter()
    if subplot is not None:
        PLOTTER.subplot(*subplot)

    # Handle element types not supported by pyvista
    if u.ufl_element().family_name not in [
        "Discontinuous Lagrange",
        "Lagrange",
        "DQ",
        "Q",
        "DP",
        "P",
    ]:
        u = _interpolate_vector_DG(u)
    if type(warp) is not bool and warp.function_space != u.function_space:
        warp = _interpolate(warp, u.function_space)

    # Gather data onto MPI process 0
    grid = _gather_grid(u.function_space.mesh, u, warp=warp, name=name)

    # Plot outline
    gdim = u.function_space.mesh.geometry.dim
    if gdim == 2 and not show_mesh and warp is False:
        outline_meshtags = _get_outline_meshtags(u.function_space.mesh)
        plot_meshtags(
            outline_meshtags,
            mesh=u.function_space.mesh,
            tagvalue=1,
            warp=warp,
            plotter=plotter,
            color="black",
            line_width=1,
            show_scalar_bar=False,
        )

    if MPI.COMM_WORLD.rank != 0:
        return PLOTTER

    # Set the global plotter
    PLOTTER = plotter if plotter is not None else PLOTTER

    # Manipulate grid data
    factor = kwargs.pop("factor") if "factor" in kwargs.keys() else 1
    glyphs = grid.glyph(orient=name if name else u.name, factor=factor)
    glyphs.rename_array("GlyphScale", name if name else u.name)

    # Default plotting options
    defaults = {"show_scalar_bar": True}

    # Visualize
    options = defaults | kwargs
    PLOTTER.add_mesh(glyphs, **options)  # Main operation
    if show_mesh and gdim in [2, 3]:
        _plot_feature_edges(PLOTTER, grid)
    if gdim in [1, 2]:
        PLOTTER.view_xy()

    if show:
        _show()

    return PLOTTER


def plot_mesh(
    mesh: dolfinx.mesh.Mesh,
    warp: dolfinx.fem.Function | bool = False,
    plotter: pyvista.Plotter | None = None,
    clear_plotter: bool = False,
    subplot: tuple | None = None,
    show: bool = False,
    **kwargs,
) -> Plotter:
    """
    Produce a plot of the mesh.

    Parameters
    ----------
    mesh : dolfinx.mesh.Mesh
        Mesh to be plotted.
    warp: dolfinx.fem.Function | bool, optional
        Field by which to warp the mesh, by default False
    plotter: Plotter | None, optional
        The Plotting object to which to add this data. If None, the default
        (module-level) instance is used, by default None
    clear_plotter : bool, optional
        Whether to clear/reset the default Plotter instance, by default
        False
    subplot : tuple | None, optional
        If the Plotter is initialized with multiple subplots, specify with
        a tuple which sublot to plot in, by default None
    show : bool, optional
        Whether to show upon completion. Alternatively, call `.show()` on
        the returned Plotter `instance`. By default False
    **kwargs
        The remaining keyword arguments are passed to the
        `plotter.add_mesh` call.

    Returns
    -------
    Plotter
        An mpi-safe child of pyvista.Plotter.
    """
    global PLOTTER
    if clear_plotter:
        reset_plotter()
    if subplot is not None:
        PLOTTER.subplot(*subplot)

    # Gather data onto MPI process 0
    grid = _gather_grid(mesh, warp=warp)
    if MPI.COMM_WORLD.rank != 0:
        return PLOTTER

    # Set the global plotter
    PLOTTER = plotter if plotter is not None else PLOTTER

    # Default plotting options
    defaults = {"style": "wireframe", "color": "black"}

    # Visualize
    options = defaults | kwargs
    _plot_feature_edges(PLOTTER, grid, options=options)
    if mesh.geometry.dim in [1, 2]:
        PLOTTER.view_xy()

    if show:
        _show()

    return PLOTTER


def plot_meshtags(
    meshtags: dolfinx.mesh.MeshTags,
    mesh: dolfinx.mesh.Mesh | None = None,
    tagvalue: int | None = None,
    warp: None = None,
    show_mesh: bool = True,
    name: str = "Tags",
    plotter: pyvista.Plotter | None = None,
    clear_plotter: bool = False,
    subplot: tuple | None = None,
    show: bool = False,
    **kwargs,
) -> Plotter:
    """
    Produce a plot of the meshtags.

    Parameters
    ----------
    meshtags:  dolfinx.mesh.MeshTags
        Meshtags to be plotted
    mesh : dolfinx.mesh.Mesh
        Mesh corresponding to the meshtags.
    tagvalue: int | None, optional
        Which tag to show. If None, will show all tags. By detault None
    warp: dolfinx.fem.Function | bool, optional
        Field by which to warp the mesh, by default False
    show_mesh: bool, optional
        Whether to overlay a plot of the mesh edges.
    name : str
        The name to give tags in the colorbar.
    plotter: Plotter | None, optional
        The Plotting object to which to add this data. If None, the default
        (module-level) instance is used, by default None
    clear_plotter : bool, optional
        Whether to clear/reset the default Plotter instance, by default
        False
    subplot : tuple | None, optional
        If the Plotter is initialized with multiple subplots, specify with
        a tuple which sublot to plot in, by default None
    show : bool, optional
        Whether to show upon completion. Alternatively, call `.show()` on
        the returned Plotter `instance`. By default False
    **kwargs
        The remaining keyword arguments are passed to the
        `plotter.add_mesh` call.

    Returns
    -------
    Plotter
        An mpi-safe child of pyvista.Plotter.
    """
    global PLOTTER
    if clear_plotter:
        PLOTTER = Plotter()
    if subplot is not None:
        PLOTTER.subplot(*subplot)

    if mesh is None:
        raise ValueError(
            "A `mesh` must be supplied to plot meshtags. See help `plot_meshtags`."
        )
    if warp:
        raise NotImplementedError("Warping a meshtags object is not supported")

    meshtags_indices, meshtag_values = _gather_meshtags(meshtags, mesh, tag=tagvalue)
    grid = _gather_grid(mesh, dim=meshtags.dim, entities=meshtags_indices)

    # Plot outline
    gdim = mesh.geometry.dim
    if gdim == 2 and not show_mesh:
        outline_meshtags = _get_outline_meshtags(mesh)
        plot_meshtags(
            outline_meshtags,
            mesh=mesh,
            tagvalue=1,
            warp=warp,
            plotter=plotter,
            color="black",
            line_width=1,
            show_scalar_bar=False,
        )

    if MPI.COMM_WORLD.rank != 0:
        return PLOTTER

    # Set the global plotter
    PLOTTER = plotter if plotter is not None else PLOTTER

    # Manipulate grid data
    grid.cell_data[name] = meshtag_values.astype(str)

    # Set plotting options:
    defaults = {"scalars": name, "show_edges": False, "show_scalar_bar": True}
    if meshtags.dim == 1:
        defaults |= {"line_width": 4}

    # Visualize
    options = defaults | kwargs
    gdim = mesh.geometry.dim
    PLOTTER.add_mesh(grid, **options)
    if show_mesh and gdim in [2, 3]:
        _plot_feature_edges(PLOTTER, grid)
    if gdim in [1, 2]:
        PLOTTER.view_xy()

    if show:
        _show()

    return PLOTTER


def plot_measure(
    measure: ufl.measure.Measure,
    mesh: dolfinx.mesh.Mesh | None = None,
    show_mesh: bool = False,
    name: str = "Measure",
    plotter: pyvista.Plotter | None = None,
    clear_plotter: bool = False,
    subplot: tuple | None = None,
    show: bool = False,
    **kwargs,
) -> Plotter:
    """
    Produce a plot of the measure, bu plotting the underlying meshtags.

    Parameters
    ----------
    measure: ufl.measure.Measure
        Measure to be plotted
    mesh : dolfinx.mesh.Mesh
        Mesh corresponding to the measure.
    show_mesh: bool, optional
        Whether to overlay a plot of the mesh edges.
    name : str
        The name to give measure in the colorbar.
    plotter: Plotter | None, optional
        The Plotting object to which to add this data. If None, the default
        (module-level) instance is used, by default None
    clear_plotter : bool, optional
        Whether to clear/reset the default Plotter instance, by default
        False
    subplot : tuple | None, optional
        If the Plotter is initialized with multiple subplots, specify with
        a tuple which sublot to plot in, by default None
    show : bool, optional
        Whether to show upon completion. Alternatively, call `.show()` on
        the returned Plotter `instance`. By default False
    **kwargs
        The remaining keyword arguments are passed to the
        `plotter.add_mesh` call.

    Returns
    -------
    Plotter
        An mpi-safe child of pyvista.Plotter.
    """
    # Forwards to plot_meshtags

    if mesh is None:
        raise ValueError(
            "A `mesh` must be supplied to plot measures. See help `plot_measure`."
        )

    meshtags = _get_measure_meshtags(measure, mesh)
    tagvalue = (
        measure.subdomain_id() if not measure.subdomain_id() == "everywhere" else None
    )
    kwargs = {"show_edges": False} | kwargs
    return plot_meshtags(
        meshtags,
        mesh=mesh,
        tagvalue=tagvalue,
        show_mesh=show_mesh,
        name=name,
        plotter=plotter,
        show=show,
        clear_plotter=clear_plotter,
        subplot=subplot,
        **kwargs,
    )


def _plot_feature_edges(
    plotter: pyvista.Plotter,
    grid: pyvista.UnstructuredGrid,
    options: dict = {"show_scalar_bar": False, "color": "black"},
    subdivisions: int = 4,
):
    surface = grid.separate_cells().extract_surface(nonlinear_subdivision=subdivisions)
    edges = surface.extract_feature_edges()
    plotter.add_mesh(edges, **options)


def _get_outline_meshtags(
    mesh: dolfinx.mesh.Mesh,
):
    tdim = mesh.topology.dim
    mesh.topology.create_entities(tdim - 1)
    mesh.topology.create_connectivity(tdim - 1, tdim)
    boundary_facets = dolfinx.mesh.exterior_facet_indices(mesh.topology)
    num_cells = mesh.topology.index_map(tdim - 1).size_local
    indices = np.arange(num_cells)
    values = np.zeros(num_cells, dtype=int)
    values[boundary_facets] = 1
    meshtags = dolfinx.mesh.meshtags(
        mesh,
        mesh.topology.dim - 1,
        indices,
        values,
    )
    return meshtags


def _get_partitioning_meshtags(mesh: dolfinx.mesh.Mesh) -> dolfinx.mesh.MeshTags:
    """
    Creates an element meshtags filled with the MPI ranks associated to
    that element
    """
    num_cells = mesh.topology.index_map(mesh.topology.dim).size_local
    meshtags = dolfinx.mesh.meshtags(
        mesh,
        mesh.topology.dim,
        np.arange(num_cells),
        np.ones(num_cells, dtype=int) * MPI.COMM_WORLD.rank,
    )
    return meshtags


def _get_measure_meshtags(
    measure: ufl.Measure, mesh: dolfinx.mesh.Mesh
) -> dolfinx.mesh.MeshTags:
    """
    Obtains the meshtags associated to an integration measure. Its values
    refer to the different integrable sub-measures.
    """
    if type(measure.subdomain_data()) is dolfinx.mesh.MeshTags:
        return measure.subdomain_data()
    num_cells = mesh.topology.index_map(mesh.topology.dim).size_local
    meshtags = dolfinx.mesh.meshtags(
        mesh, mesh.topology.dim, np.arange(num_cells), np.ones(num_cells)
    )
    return meshtags


def _interpolate_vector_DG(u: dolfinx.fem.Function):
    """
    Interpolate a vector-values field onto a DG superspace. For
    visualization of advanced FE spaces.
    """
    V = u.function_space
    domain = V.mesh
    gdim = domain.geometry.dim
    poldeg = V.element.basix_element.degree
    VDG = dolfinx.fem.functionspace(domain, ("Discontinuous Lagrange", poldeg, (gdim,)))
    return _interpolate(u, VDG)


def _interpolate(u: dolfinx.fem.Function, V: dolfinx.fem.FunctionSpace):
    """
    Basic interpolation helper.
    """
    u_new = dolfinx.fem.Function(V, name=u.name)
    u_new.interpolate(u)
    return u_new


def _compatible_scalar_u_warp(
    u: dolfinx.fem.Function, warp: dolfinx.fem.Function | bool
):
    """
    Ensures that the scalar u field and the warp field can be plotted on
    the same mesh. Essentially, the warp field is projected onto the
    functionspace of a vectorized u
    """
    V = u.function_space
    domain = V.mesh
    if V.element.basix_element.degree == 0:
        # DG0 functionspace can't be passed to plot.vtk_mesh
        V = dolfinx.fem.functionspace(domain, ("Discontinuous Lagrange", 1))
        u = _interpolate(u, V)

    if type(warp) == bool:
        return u, warp

    family_name = u.ufl_element().family_name
    discontinuous = u.ufl_element().discontinuous
    poldeg = V.element.basix_element.degree
    gdim = domain.geometry.dim

    if (
        warp.ufl_element().family_name != family_name
        or warp.function_space.element.basix_element.degree != poldeg
        or warp.ufl_element().discontinuous != discontinuous
    ):
        el = basix.ufl.element(
            family_name,
            domain.ufl_cell().cellname(),
            poldeg,
            shape=(gdim,),
            discontinuous=discontinuous,
        )
        Vwarp = dolfinx.fem.functionspace(domain, el)
        warp = _interpolate(warp, Vwarp)

    return u, warp
